module ama;

// 条件設定プラグイン
// 2012.6.5 飯香


import std.stdio;
import std.file;
import std.conv;
import std.string;
import std.path;
import std.c.windows.windows;
import std.windows.charset;

import utauio;
import kagonote;
import voicebank;
import otoini;

import accsetting;
import accconv;
import acclog;

Log log, outlog;

// メイン ============================================================

int main(string[] args){
	
	SettingData settingData;
	InoutData inoutData;
	Voicebank voicebank;
	
	// ログ開始
	log = new Log("ama.log");
	log.add("ama");
	log.add(date.toString(date.UTCtoLocalTime(date.getUTCtime())));
	log.add("");
	
	SettingData.log = log;
	otoini.log = log;
	prefixmap.log = log;
	Voicebank.log = log;
	Note.log = log;
	
	outlog = new Log("amaout.log");
	
	// ログは終了時に閉じる
	scope(exit) log.output(), outlog.output();
	scope(success) log.add("Successful Program Termination.");
	scope(failure) log.add("Unsuccessful Program Termination. ");
	
	// 実行ファイルのファイル名
	// :: これLogのコンストラクタに持ってく？ または何かそういうオブジェクト？
	string commandLine;
	string exeFileName, exeFileDir, szExeFileName, szExeFileDir;
	LPSTR lpFileName = cast(char *) new char[MAX_PATH];
	GetModuleFileNameA(null, lpFileName, MAX_PATH);
	szExeFileName = to!string(lpFileName);
	exeFileName = fromMBSz(cast(immutable(char)*)szExeFileName);
	exeFileDir = dirname(exeFileName);
	szExeFileDir = to!string(toMBSz(exeFileDir));
	log.add("Execute file name: " ~ szExeFileName);
	log.add("Execute file dir: " ~ szExeFileDir);
	
	// UTAUとのやりとり一時ファイルの名前を取得
	string inoutFileName;
	if(args.length>1) inoutFileName = to!string(std.windows.charset.toMBSz(args[1]));
	else log.err("Can't read command line.");
	log.add("Temporary file name: " ~ inoutFileName);
	
	// UTAUとのやりとり一時ファイルを読み込む
	// InoutData オブジェクトを作成
	try inoutData = new InoutData(fromMBSz(cast(immutable(char)*)toStringz(inoutFileName)));
	catch(Exception e) log.err("Failed to open temporary file. - " ~ to!string(std.windows.charset.toMBSz(e.toString())));
	log.add("Temporary file read.");
	
	// 篭の設定ファイルを読み込み
	try settingData = new SettingData(exeFileDir ~ "\\settings.txt");
	catch(Exception e) log.add("Failed to open setting file. - " ~ to!string(std.windows.charset.toMBSz(e.toString())));
	Voice.settingData = settingData;
	OtoIni.settingData = settingData;
	
	// 音源を読み込み
	if(inoutData.settingSession && "VoiceDir" in inoutData.settingSession.attributes){
		log.add("voicebank path: " ~ inoutData.settingSession.attributes["VoiceDir"]);
		try voicebank = new Voicebank(sjis2utf8(inoutData.settingSession.attributes["VoiceDir"]));
		catch(Exception e) log.err("Failed to open voicebank. - " ~ to!string(std.windows.charset.toMBSz(e.toString())));
		}
	
	// 亜麻の設定ファイルを読み込み
	// 音の集合
	AmaSetting ifSetting;
	try ifSetting = new AmaSetting(exeFileDir ~ "\\ama-if.txt");
	catch(Exception e) log.add("Failed to open setting file. - " ~ to!string(std.windows.charset.toMBSz(e.toString())));
	// 設定値(pre, ovl, stp)
	AmaSetting thenSetting;
	try thenSetting = new AmaSetting(exeFileDir ~ "\\ama-then.txt");
	catch(Exception e) log.add("Failed to open setting file. - " ~ to!string(std.windows.charset.toMBSz(e.toString())));
	
	// セッションを読み込んで条件を適用する
	foreach(Session ses; inoutData.originalSessions){
		
		// セッションに対応するノート
		AmaNote n =new AmaNote(voicebank);
		
		// ノートからセッションへの参照
		n.session = ses;
		
		// ノートに属性値などを設定
		if("Lyric" in ses.attributes) n.lyric = ses.attributes["Lyric"];
		if("NoteNum" in ses.attributes)
				n.notenum = to!int(to!float(ses.attributes["NoteNum"]));
		if("Length" in ses.attributes)
				n.duration = to!int(to!float(ses.attributes["Length"]));
		if("PreUtterance" in ses.attributes && ses.attributes["PreUtterance"].length)
				n.preUtterance = to!int(to!float(ses.attributes["PreUtterance"]));
		if("VoiceOverlap" in ses.attributes && ses.attributes["VoiceOverlap"].length)
				n.voiceOverlap = to!int(to!float(ses.attributes["VoiceOverlap"]));
		if("stp" in ses.attributes && ses.attributes["stp"].length)
				n.stp = to!int(to!float(ses.attributes["stp"]));
		if("StartPoint" in ses.attributes && ses.attributes["StartPoint"].length)
				n.stp = to!int(to!float(ses.attributes["StartPoint"]));
		if("StartPoint" in ses.attributes && ses.attributes["StartPoint"].length)
				n.stp = to!int(to!float(ses.attributes["StartPoint"]));
		if("Moduration" in ses.attributes)
				n.moduration = to!int(to!float(ses.attributes["Moduration"]));
		if("Intensity" in ses.attributes)
				n.intensity = to!int(to!float(ses.attributes["Intensity"]));
		
		log.add("");
		log.add("\"" ~ n.lyric ~ "\"");
		
		// どの条件名に該当するか調べる
		foreach(SettingNomination nom; ifSetting.settings){
			log.add("");
			if(nom.test(n)){
				n.settingname = nom.name;
				log.add("hit!");
				log.add("Applying [" ~ nom.name ~ "]...");
				
				// 条件名に対応した設定を適用
				foreach(SettingNomination thennom; thenSetting.settings){
					if(thennom.name == n.settingname) thennom.apply(n);
					}
				
				break;
				}
			log.add("failed.");
			}
		}
	
	// InoutData を書き出し
	// （AmaNote.session は参照なのですでに反映されている）
	inoutData.output();
	
	outlog.add(inoutData.dump);
	return 0;
	}


// 条件定義ファイル（if定義とthen定義を兼用）
class AmaSetting{
	SettingNomination[] settings;
	this(string filename){
		log.add("");
		log.addUtf("loading setting from: " ~ filename);
		string[] lines = splitlines(cast(string)read(filename));
		SettingNomination currentSetting;
		foreach(string l; lines){
			log.add("");
			if(l.length>1 && l[0..1] == "[" && l[length-1..length] == "]"){
				settings ~= new SettingNomination(l[1..length-1]);
				currentSetting = settings[length-1];
				log.add("[" ~ l[1..length-1] ~ "] nomination");
				}
			else if(currentSetting && l.length>0){
				log.add(l);
				currentSetting.lines ~= new SettingLine(l);
				}
			}
		}
	}

// 条件名
class SettingNomination{
	string name;
	SettingLine[] lines;
	this(string n){
		this.name = n;
		}
	bool test(AmaNote n){
		log.add("[" ~ name ~ "]");
		if(!lines.length) return false;
		foreach(SettingLine l; lines){
			if(!l.test(n)) return false;
			}
		return true;
		}
	void apply(AmaNote n){
		foreach(SettingLine l; lines) l.apply(n);
		}
	}

// 定義ファイルの行
class SettingLine{
	string objectname;
	string operator;
	string[] values;
	this(string line){
		string[] columns = [];
		string k = "", j = "", s = "";
		int i=0;
		for( ; i<line.length; i++){
			if(line[i] != '=') k ~= line[i];
			else break;
			}
		for(i++; i<line.length; i++){
			if(line[i] != ',') s ~= line[i];
			else{
				columns ~= s;
				s = "";
				}
			}
		if(k.length) switch(k[length-1..length]){
			case "<": case ">": case "!": case "+": case "-": case "*":
				j = k[length-1..length];
				k = k[0..length-1];
			default:
			}
		if(s.length) columns ~= s;
		log.add("objectname=" ~ k ~ " operator=" ~ j ~ " values=" ~ columns.join(" "));
		objectname = k;
		operator = j;
		values = columns;
		}
	// ノートが該当するかどうか検査
	bool test(AmaNote n){
		log.clear;
		// 左辺値を取得
		string left;
		switch(objectname){
			case "lyric":
				left = n.noteLyric;
				break;
			case "baselyric":
				left = n.baseLyric;
				break;
			case "prevowel":
				if(n.prevowel.length) left = [n.prevowel[0]];
				else left = "";
				break;
			case "keyname":
				left = n.keyname;
				break;
			case "notenum":
				left = to!string(n.notenum);
				break;
			case "duration":
				left = to!string(n.duration);
				break;
			case "moduration":
				left = to!string(n.moduration);
				break;
			case "intensity":
				left = to!string(n.intensity);
				break;
			case "preutterance":
				left = to!string(n.preUtterance);
				break;
			case "voiceoverlap":
				left = to!string(n.voiceOverlap);
				break;
			case "stp":
				left = to!string(n.stp);
				break;
			case "preutteranceR":
				left = to!string(n.preUtteranceRelative);
				break;
			case "voiceoverlapR":
				left = to!string(n.voiceOverlapRelative);
				break;
			case "stpR":
				left = to!string(n.stpRelative);
				break;
			default:
				if(objectname.length && objectname[0] == '#'){
					if(objectname[1..$] in n.session.flags) left = n.session.flags[objectname[1..$]];
					else left = "";
					}
				else if(objectname in n.session.attributes) left = n.session.attributes[objectname];
				else left = "";
				break;
			}
		log.addInline("testing: " ~ left);
		// 演算子によって分岐
		switch(operator){
			case "":
				log.addInline("=" ~ values.join(" or ") );
				if(left == "" && values.length == 0) return true;
				foreach(string v; values) if(left == v) return true;
				return false;
			case ">":
				log.addInline(">=" ~ values.join(" or ") );
				if(!isNumeric(left)) left = "0";
				foreach(string v; values) if(to!float(left) >= to!float(v)) return true;
				return false;
			case "<":
				log.addInline("<=" ~ values.join(" or ") );
				if(!isNumeric(left)) left = "0";
				foreach(string v; values) if(to!float(left) <= to!float(v)) return true;
				return false;
			case "!":
				log.addInline("!=" ~ values.join(" or ") );
				if(left == "" && values.length == 0) return false;
				foreach(string v; values) if(left == v) return false;
				return true;
			default:
				log.add("Illegal if-line:" ~ dump);
				return false;
			}
		}
	// ノートに設定を適用
	void apply(AmaNote n){
		string value;
		if(!values.length) value ~= "";
		else value = values[0];
		log.add(dump);
		switch(objectname){
			case "lyric":
				n.prevowel = "";
			case "baselyric":
				n.lyric = value;
				n.session.attributes["Lyric"] = n.noteLyric;
				break;
			case "prevowel":
				n.prevowel = value;
				n.session.attributes["Lyric"] = n.noteLyric;
				break;
			case "keyname":
				if(operator == "+") n.notenum = n.notenum + to!int(to!float(value));
				else if(operator == "-") n.notenum = n.notenum - to!int(to!float(value));
				else n.keyname = value;
				n.session.attributes["NoteNum"] = to!string(n.notenum);
				break;
			case "notenum":
				if(operator == "+") n.notenum = n.notenum + to!int(to!float(value));
				else if(operator == "-") n.notenum = n.notenum - to!int(to!float(value));
				else n.notenum = to!int(to!float(value));
				n.session.attributes["NoteNum"] = to!string(n.notenum);
				break;
			case "duration":
				if(operator == "+") n.duration += to!float(value);
				else if(operator == "-") n.duration -= to!float(value);
				else n.duration = to!int(to!float(value));
				n.session.attributes["Length"] = to!string(n.duration);
				break;
			case "moduration":
				if(operator == "+") n.moduration += to!float(value);
				else if(operator == "-") n.moduration -= to!float(value);
				else n.moduration = to!int(to!float(value));
				n.session.attributes["Moduration"] = to!string(n.moduration);
				break;
			case "intensity":
				if(operator == "+") n.intensity += to!float(value);
				else if(operator == "-") n.intensity -= to!float(value);
				else n.intensity = to!int(to!float(value));
				n.session.attributes["Intensity"] = to!string(n.intensity);
				break;
			case "preutterance":
				if(operator == "+") n.preUtterance = n.preUtterance + to!int(to!float(value));
				else if(operator == "-") n.preUtterance = n.preUtterance - to!int(to!float(value));
				else n.preUtterance = to!int(to!float(value));
				n.session.attributes["PreUtterance"] = to!string(n.preUtterance);
				break;
			case "voiceoverlap":
				if(operator == "+") n.voiceOverlap = n.voiceOverlap + to!int(to!float(value));
				else if(operator == "-") n.voiceOverlap = n.voiceOverlap - to!int(to!float(value));
				else n.voiceOverlap = to!int(to!float(value));
				n.session.attributes["VoiceOverlap"] = to!string(n.voiceOverlap);
				break;
			case "stp":
				if(operator == "+") n.stp = n.stp + to!int(to!float(value));
				else if(operator == "-") n.stp = n.stp - to!int(to!float(value));
				else n.stp = to!int(to!float(value));
				n.session.attributes["stp"] = to!string(n.stp);
				n.session.attributes["StartPoint"] = to!string(n.stp);
				break;
			case "preutteranceR":
				if(operator == "+") n.preUtteranceRelative = n.preUtteranceRelative + to!int(to!float(value));
				else if(operator == "-") n.preUtteranceRelative = n.preUtteranceRelative - to!int(to!float(value));
				else n.preUtteranceRelative = to!int(to!float(value));
				n.session.attributes["PreUtterance"] = to!string(n.preUtterance);
				break;
			case "voiceoverlapR":
				if(operator == "+") n.voiceOverlapRelative = n.voiceOverlapRelative + to!int(to!float(value));
				else if(operator == "-") n.voiceOverlapRelative = n.voiceOverlapRelative - to!int(to!float(value));
				else n.voiceOverlapRelative = to!int(to!float(value));
				n.session.attributes["VoiceOverlap"] = to!string(n.voiceOverlap);
				break;
			case "stpR":
				if(operator == "+") n.stpRelative = n.stpRelative + to!int(to!float(value));
				else if(operator == "-") n.stpRelative = n.stpRelative - to!int(to!float(value));
				else n.stpRelative = to!int(to!float(value));
				n.session.attributes["stp"] = to!string(n.stp);
				n.session.attributes["StartPoint"] = to!string(n.stp);
				break;
			default:
				if(objectname.length && objectname[0] == '#'){
					if(!(objectname[1..$] in n.session.flags)) n.session.flags[objectname[1..$]] = "0";
					if(operator == "+") n.session.flags[objectname[1..$]] = to!string(to!int(to!float(n.session.flags[objectname[1..$]])) + to!int(to!float(value)));
					else if(operator == "-") n.session.flags[objectname[1..$]] = to!string(to!int(to!float(n.session.flags[objectname[1..$]])) - to!int(to!float(value)));
					else n.session.flags[objectname[1..$]] = value;
					}
				else{
					if(!(objectname in n.session.attributes)) n.session.attributes[objectname] = "0";
					if(operator == "+") n.session.attributes[objectname] = to!string(to!int(to!float(n.session.attributes[objectname])) + to!int(to!float(value)));
					else if(operator == "-") n.session.attributes[objectname] = to!string(to!int(to!float(n.session.attributes[objectname])) - to!int(to!float(value)));
					else n.session.attributes[objectname] = value;
					}
				break;
			}
		}
	string dump(){
		return objectname ~ operator ~ "=" ~ values.join(",");
		}
	}

class AmaNote: Note{
	Session session;
	string settingname;
	this(Voicebank vb){
		super(vb);
		}
	}

